from tr.common import *
import jsonschema


class VerValues(object):
    MAJOR: int = 4
    MINOR: int = 4


class VerKeys(object):
    Major = "major"
    Minor = "minor"


class _CommonKeys(object):
    Variables = "variables"


class AutoVarsKeys(object):
    TASK_ROOT = "taskRoot"
    CWD = "cwd"
    TASK_CLI_ARGS = "cliArgs"


class TaskKeys(object):
    Shell = "shell"
    ShellPath = "shell_path"
    Env = "env"
    EnvInherit = "env_inherit"
    Cwd = "cwd"
    Base = "base"
    Hidden = "hidden"
    Abstract = "abstract"
    ShortDesc = "short_desc"
    LongDesc = "description"
    Commands = "commands"
    StopOnError = "stop_on_error"
    Meta = "meta"
    Variables = _CommonKeys.Variables
    # Container settings
    CImage = "c_image"
    CTool = "c_container_tool"
    CVolumes = "c_volumes"
    CInteractive = "c_interactive"
    CTty = "c_tty"
    CFlags = "c_flags"
    CExec = "c_exec"
    CRemove = "c_remove"
    CSudo = "c_sudo"
    CShell = "c_shell"
    CShellPath = "c_shell_path"
    CEnv = "c_env"
    CCwd = "c_cwd"


class GlobalKeys(object):
    Schema = "$schema"
    Include = "include"
    UseDfltInclude = "use_default_include"
    Tasks = "tasks"
    Suppress = "suppress"
    Variables = _CommonKeys.Variables
    DfltTask = "default_task"
    DfltShellPath = "default_shell_path"
    DfltContainerShellPath = "default_container_shell_path"
    DfltContainerTool = "default_container_tool"


_task_schema = {
    "type": "object",
    "properties": {
        TaskKeys.Base: {"type": "string", "minLength": 1},
        TaskKeys.ShortDesc: {"type": "string", "maxLength": 75},
        TaskKeys.LongDesc: {"type": "string"},
        TaskKeys.Commands: {
            "type": "array",
            "items": {"type": "string", "minLength": 1}
        },
        TaskKeys.Cwd: {"type": "string", "minLength": 1},
        TaskKeys.Shell: {"type": "boolean"},
        TaskKeys.ShellPath: {"type": "string", "minLength": 1},
        TaskKeys.Env: {
            "type": "object",
            "additionalProperties": {
                "type": "string"
            }
        },
        TaskKeys.EnvInherit: {"type": "boolean"},
        TaskKeys.StopOnError: {"type": "boolean"},
        TaskKeys.Hidden: {"type": "boolean"},
        TaskKeys.Abstract: {"type": "boolean"},
        TaskKeys.Variables: {"type": "object"},
        TaskKeys.CImage: {"type": "string"},
        TaskKeys.CTool: {"type": "string"},
        TaskKeys.CVolumes: {
            "type": "array",
            "items": {"type": "string"}
        },
        TaskKeys.CInteractive: {"type": "boolean"},
        TaskKeys.CTty: {"type": "boolean"},
        TaskKeys.CRemove: {"type": "boolean"},
        TaskKeys.CFlags: {"type": "string"},
        TaskKeys.CExec: {"type": "boolean"},
        TaskKeys.CCwd: {"type": "string", "minLength": 1},
        TaskKeys.CShell: {"type": "boolean"},
        TaskKeys.CShellPath: {"type": "string", "minLength": 1},
        TaskKeys.CSudo: {"type": "boolean"},
        TaskKeys.CEnv: {
            "type": "object",
            "additionalProperties": {
                "type": "string"
            }
        },
        TaskKeys.Meta: {"type": "object"},
    },
    "additionalProperties": False
}

_config_file_schema = {
    "type": "object",
    "properties": {
        GlobalKeys.Schema: {
            "type": "string",
            "minLength": 1
        },
        GlobalKeys.Include: {
            "type": "array",
            "items": {"type": "string", "minLength": 1}
        },
        GlobalKeys.UseDfltInclude: {"type": "boolean"},

        GlobalKeys.Tasks: {"type": "object"},
        GlobalKeys.Suppress: {
            "type": "array",
            "items": {"type": "string", "minLength": 1}
        },
        GlobalKeys.Variables: {"type": "object"},
        GlobalKeys.DfltTask: {"type": "string"},
        GlobalKeys.DfltShellPath: {
            "type": "string",
            "minLength": 1
        },
        GlobalKeys.DfltContainerTool: {"type": "string", "minLength": 1},
    },
    "additionalProperties": False
}


def validate_config_file_schema(data: dict) -> None:
    try:
        jsonschema.validate(data, _config_file_schema)
    except jsonschema.ValidationError as e:
        raise TaskException("Schema validation error at '{}': {}".format(
            "/".join(str(v) for v in list(e.absolute_path)), e.message))


def validate_task_schema(name: str, data: dict) -> None:
    try:
        jsonschema.validate(data, _task_schema)
    except jsonschema.ValidationError as e:
        raise TaskException(f"Task schema validation error for '{name}': {e.message}")


class SchemaDumpOpts(object):
    ALL = "all"
    CONFIG = "config"
    TASK = "task"

    CHOICES = [ALL, CONFIG, TASK]


def dump_schema(dump_type: str) -> None:
    if dump_type == SchemaDumpOpts.CONFIG:
        print_dict(_config_file_schema)
    elif dump_type == SchemaDumpOpts.TASK:
        print_dict(_task_schema)
    elif dump_type == SchemaDumpOpts.ALL:
        _config_file_schema[GlobalKeys.Tasks] = _task_schema
        print_dict(_config_file_schema)



from tr import TR_BASE_VERSION
from tr.config import *
from tr.schemas import AutoVarsKeys
import tr.actions as actions
import argparse
import argcomplete
import sys

__RUN_CMD = "run"
__LIST_CMD = "list"
__INFO_CMD = "info"
__DUMP_TASK_CMD = "dump"
__DUMP_CONFIG_CMD = "dump_config"
__DUMP_SCHEMA_CMD = "dump_schema"


def _tasks_complete(**kwargs) -> typing.List[str]:
    try:
        parsed_args: argparse.Namespace = kwargs['parsed_args']
        parser_name = parsed_args.subparsers_name
    except (KeyError, AttributeError):
        return []
    if (parser_name == __RUN_CMD or parser_name == __INFO_CMD or parser_name == __DUMP_TASK_CMD) \
       and parsed_args.task is None:
        return Config(None).visible_tasks()
    return []


def _parse_arguments() -> argparse.Namespace:
    try:
        ex_args_idx = sys.argv.index('--')
        tr_argv = sys.argv[1:ex_args_idx]
        cmds_argv = sys.argv[ex_args_idx + 1:]
    except ValueError:
        tr_argv = sys.argv[1:]
        cmds_argv = []
    yes_no: typing.List[str] = [TASK_YES_TOKEN, TASK_NO_TOKEN]
    parser = argparse.ArgumentParser(prog='task')
    parser.add_argument('--version', action='version', version=f'%(prog)s {TR_BASE_VERSION}')
    parser.add_argument('-v', '--verbose', action='count', help='log file verbosity', default=0)
    parser.add_argument('-C', '--conf', metavar='CONF', help='configuration file to use',
                        default=None)
    parser.add_argument('--log_file', metavar='FILE', help='set log file', default='')
    subparsers = parser.add_subparsers(help='commands', dest='subparsers_name')
    subparsers.required = True

    task_target_parser = argparse.ArgumentParser(add_help=False)
    task_target_parser.add_argument('task', nargs='?', metavar='TASK', default=None,
                                    help='set task')
    task_target_parser.add_argument('-V', '--variable', metavar='VAR', default=[], action='append',
                                    help='set a variable')
    run_parser = subparsers.add_parser(__RUN_CMD, help='execute a task',
                                       parents=[task_target_parser])
    run_parser.add_argument('-c', '--command', metavar='CMD', default=None, action='append',
                            help='set command to run')
    run_parser.add_argument('--cwd', metavar='DIR', default=None, help='set working directory')
    run_parser.add_argument('--shell', type=str, choices=yes_no, action='store', default=None,
                            help='set shell usage')
    run_parser.add_argument('--shell-path', metavar='PATH', help='set shell path', default=None)
    run_parser.add_argument('--stop-on-error', choices=yes_no, action='store', default=None,
                            help='set stop behavior on command error')
    run_parser.add_argument('--env', metavar='ENV=VAL', default=None, action='append',
                            help='set an environment variable')
    run_parser.add_argument('-s', '--summary', action='store_true', default=False,
                            help='show task summary before run')

    #  Container specific arguments
    run_parser.add_argument('--c-image', metavar='IMAGE',
                            help='set image or container to run the task in')
    run_parser.add_argument('--c-tool', metavar='TOOL',
                            help='set container tool to use')
    run_parser.add_argument('--c-volume', metavar='VOLUME', action='append', default=None,
                            help='set container volume')
    run_parser.add_argument('--c-tty', choices=yes_no, action='store', default=None,
                            help='set container tty allocation')
    run_parser.add_argument('--c-interactive', choices=yes_no, action='store', default=None,
                            help='set container interactive mode')

    cont_run_type_grp = run_parser.add_mutually_exclusive_group()
    cont_run_type_grp.add_argument('--c-rm', type=str, choices=yes_no, action='store', default=None,
                                   help='set container removal after run')
    cont_run_type_grp.add_argument('--c-exec', action='store_true', default=False,
                                   help='run command in existing container')
    run_parser.add_argument('--c-flags', metavar='FLAGS', default=None,
                            help='set container flags')
    run_parser.add_argument('--c-shell', type=str, choices=yes_no, action='store', default=None,
                            help='wrap container command in shell')
    run_parser.add_argument('--c-shell-path', metavar='PATH', default=None,
                            help='set container shell path')
    run_parser.add_argument('--c-env', metavar='ENV=VAL', default=None, action='append',
                            help='set container environment variable')
    run_parser.add_argument('--c-cwd', metavar='DIR', default=None,
                            help='set container working directory')

    info_parser = subparsers.add_parser(__INFO_CMD, help='show task info',
                                        parents=[task_target_parser])
    info_parser.add_argument('-x', '--expand', help='expand values', action='store_true',
                             default=False)

    dump_parser = subparsers.add_parser(__DUMP_TASK_CMD, help='dump a task',
                                        parents=[task_target_parser])
    dump_parser.add_argument('-i', '--includes', help='with inclusions',
                             action='store_true', default=False)

    list_parser = subparsers.add_parser(__LIST_CMD, help='list tasks')
    list_parser.add_argument('-a', '--all', action='store_true', default=False,
                             help='show hidden and shadowed tasks')
    list_parser.add_argument('--names-only', action='store_true', default=False,
                             help=argparse.SUPPRESS)

    dump_parser = subparsers.add_parser(__DUMP_SCHEMA_CMD, help='dump configuration file schema')
    dump_parser.add_argument('-t', '--type', choices=SchemaDumpOpts.CHOICES,
                             default=SchemaDumpOpts.ALL)

    subparsers.add_parser(__DUMP_CONFIG_CMD, help='dump configuration')

    # TODO: Not sure what pyright wants with this type ignore
    argcomplete.autocomplete(parser, always_complete_options=False,
                             default_completer=_tasks_complete)  # type: ignore
    try:
        args = parser.parse_args(tr_argv)
    except SystemExit:
        raise TaskException("")
    args.__setattr__(AutoVarsKeys.TASK_CLI_ARGS, cmds_argv)
    return args


def task_runner_main() -> int:
    try:
        args = _parse_arguments()
        init_logging(args.log_file, args.verbose)

        info("args='{}'", sys.argv[1:])
        info("cmd_args={}", args.__getattribute__(AutoVarsKeys.TASK_CLI_ARGS))

        if args.subparsers_name == __DUMP_SCHEMA_CMD:
            dump_schema(args.type)
            return 0

        config = Config(args)
        if args.subparsers_name == __RUN_CMD:
            return actions.run_task(config)
        elif args.subparsers_name == __LIST_CMD:
            actions.list_tasks(config)
        elif args.subparsers_name == __INFO_CMD:
            actions.show_task_info(config)
        elif args.subparsers_name == __DUMP_CONFIG_CMD:
            actions.dump_config(config)
        elif args.subparsers_name == __DUMP_TASK_CMD:
            actions.dump_task(config)

    except TaskException as e:
        error_and_print(str(e))
        return 255
    return 0



from tr.Task import *
import textwrap
from argparse import Namespace as Args


def _active_task_name(config: Config) -> str:
    if config.args and config.args.task:
        task = config.args.task
    else:
        task = config.default_task_name()
    if task is None:
        raise TaskException("No main task name was given")
    return task


def _show_task(task: Task, full_details: bool) -> None:
    def print_val(title: str, value: typing.Any) -> None:
        print(f"{title:<24}{value:<}")

    def print_bool(title, value: bool) -> None:
        print_val(title, "Yes" if value else "No")

    def print_blob(title: str, text: str) -> None:
        if text is None or len(text) == 0:
            print(title)
            return
        first = True
        for line in text.split('\n'):
            for in_line in textwrap.wrap(line, width=80):

                if first:
                    print_val(title, in_line)
                    first = False
                    continue
                print_val("", in_line)

    def _task_str(cmd_str: str) -> str:
        if len(cmd_str.strip()) == 0:
            return "[n/a - empty string)]"
        return cmd_str

    print_val("Task name:", task.name)
    if full_details:
        if task.short_desc:
            print_val("Short description:", task.short_desc)
        if task.long_desc:
            print_blob("Description:", task.long_desc)
        print_bool("Hidden:", task.hidden)
        print_bool("Abstract:", task.abstract)
    print_bool("Use shell: ", task.shell)
    if task.shell:
        shell_title = "Shell path:"
        if task.shell_path is None:
            print_val(shell_title, "/usr/bin/sh")
        else:
            print_val(shell_title, task.shell_path)
    print_bool("Inherit environment", task.env_inherit)
    count = 0
    if task.env:
        print("Environment:")
        for k, v in task.env.items():
            print_blob(f"     [{count}]", f"{k}={v}")
            count += 1
    if task.cwd:
        print_blob("Working directory:", _task_str(task.cwd))

    if task.c_image:
        print("Container details:")
        image = _task_str(task.c_image)
        if task.c_exec:
            print_val("  Execute in:", image)
        else:
            print_val("  Run image:", image)
            print_bool("  Remove:", task.c_rm)
        print_bool("  Interactive:", task.c_interactive)
        print_bool("  Allocate tty:", task.c_tty)
        print_bool("  Use shell: ", task.c_shell)
        if task.c_shell:
            print_val("  Shell path:", task.c_shell_path)
        if task.c_flags:
            print_blob("  Run/Exec flags:", task.c_flags)
        if task.c_cwd:
            print_blob("  Working directory:", _task_str(task.c_cwd))
        if len(task.c_volumes) == 1:
            print_blob("  Volume:", _task_str(task.c_volumes[0]))
        elif len(task.c_volumes) > 1:
            count = 0
            print("  Volumes:")
            for vol in task.c_volumes:
                print_blob(f"       [{count}]", _task_str(vol))
                count += 1
        if task.c_env:
            print("  Environment:")
            for k, v in task.c_env.items():
                print_blob(f"       [{count}]", f"{k}={v}")
                count += 1

    if len(task.commands) == 0:
        if task.c_image is None:
            print("NOTICE: No commands defined for task")
        else:
            print("NOTICE: Will run the image/container default command")
        return
    if len(task.commands) == 1:
        print_blob("Command:", _task_str(task.commands[0]))
        return

    print_bool("Stop on error:", task.stop_on_error)
    print("Commands:")
    count = 0
    for cmd in task.commands:
        print_blob(f"     [{count}]", _task_str(cmd))
        count += 1


def show_task_info(config: Config) -> None:
    task_name = _active_task_name(config)
    task = Task(task_name, config)
    if config.args.expand:
        task.expand()
    _show_task(task, True)


def list_tasks(config: Config) -> None:
    show_all: bool = config.args.all
    names_only: bool = config.args.names_only
    info("Listing tasks show_all={} names_only={}", show_all, names_only)
    print_fmt = "{:<24}{:<6}{}"
    err_print_fmt = "{:<30}{}"
    default_task_name = config.default_task_name()

    if not names_only:
        print(print_fmt.format("Name", "Flags", "Description"))
        print(print_fmt.format("----", "-----", "-----------"))
    for task_name in sorted([*config.tasks]):
        try:
            t = Task(task_name, config=config)
        except TaskException as e:
            print(err_print_fmt.format(task_name, f"<error: {e}>"))
            continue
        if not show_all and (t.hidden or t.abstract):
            continue
        if names_only:
            print(task_name, end=' ')
            continue

        if t.short_desc:
            desc = t.short_desc if len(t.short_desc) <= 55 else t.short_desc[:52] + "..."
        else:
            desc = ""

        flags = ""
        if t.abstract:
            flags += "A"
        elif t.hidden:
            flags += "H"
        elif task_name == default_task_name:
            flags += "*"

        print(print_fmt.format(task_name, flags, desc[-55:]))


def args_update(task, args: Args) -> None:
    if args.stop_on_error:
        task.stop_on_error = args.stop_on_error
    if args.command:
        task.commands = args.command
    if args.cwd:
        task.cwd = args.cwd
    if args.shell:
        task.shell = (args.shell == TASK_YES_TOKEN)
    if args.shell_path:
        task.shell_path = args.shell_path
    if args.env:
        task.env = {}
        for e in args.env:
            e_name, e_value = parse_assignment_str(e)
            task.env[e_name] = e_value

    if args.c_image:
        task.c_image = args.c_image
    if args.c_volume:
        task.c_volumes = args.c_volume
    if args.c_interactive:
        task.c_interactive = (args.c_interactive == TASK_YES_TOKEN)
    if args.c_tty:
        task.c_tty = (args.c_tty == TASK_YES_TOKEN)
    if args.c_flags:
        task.c_flags = args.c_flags
    if args.c_exec:
        task.c_exec = args.c_exec
    if args.c_rm:
        task.c_rm = (args.c_rm == TASK_YES_TOKEN)
    if args.c_tool:
        task.c_tool = args.c_tool
    if args.c_shell:
        task.c_shell = (args.c_shell == TASK_YES_TOKEN)
    if args.c_shell_path:
        task.c_shell_path = args.c_shell_path
    if args.c_cwd:
        task.c_cwd = args.c_cwd
    if args.c_env:
        task.c_env = {}
        for e in args.c_env:
            e_name, e_value = parse_assignment_str(e)
            task.c_env[e_name] = e_value
    for v in args.variable:
        key, val = parse_assignment_str(v)
        task.vars_map[key] = val


def run_task(config: Config) -> int:
    task_name = _active_task_name(config)
    info("Running task '{}'", task_name)
    task = Task(task_name, config)
    args_update(task, config.args)
    task.expand()
    if config.args.summary:
        _show_task(task, False)
        print("-" * 70)
        sys.stdout.flush()
    return task.run()


def dump_task(config: Config) -> None:
    task_name = _active_task_name(config)
    print(json.dumps(config.get_task_desc(task_name, config.args.includes), indent=4))


def dump_config(config: Config) -> None:
    print(json.dumps(config.conf, indent=4))



_MAJOR_VER = 0
_MINOR_VER = 7
_FIX_VER = 1
_BUILD_VER = 1
TR_BASE_VERSION = f'{_MAJOR_VER}.{_MINOR_VER}.{_FIX_VER}'
TR_FULL_VERSION = f'{_MAJOR_VER}.{_MINOR_VER}.{_FIX_VER}-{_BUILD_VER}'
__VERSION__ = TR_FULL_VERSION



from tr.logTools import *
import sys
import re
import typing
import os
import traceback
import json
import copy
from typing import Optional

TASK_YES_TOKEN = 'yes'
TASK_NO_TOKEN = 'no'


class TaskException(Exception):
    def __init__(self, error: str) -> None:
        self.error = error

    def __str__(self) -> str:
        return self.error


def _dict_path_elements(path: str) -> list:
    elements = []
    for element in path.split('/'):
        if element.startswith("#"):
            try:
                index_el = int(element[1:])
                element = index_el
            except (IndexError, ValueError):
                pass
        elements.append(element)
    return elements


def dict_value(d: dict, path: str, require=False, default=None) -> typing.Any:
    elements = _dict_path_elements(path)
    field = d
    try:
        for element in elements:
            field = field[element]
        return field

    except (KeyError, IndexError):
        if require:
            raise TaskException(f"No '{path}' setting was found")
    return default


_const_vars_map: dict = {}
_global_vars_map: dict = {}
_default_vars_map: dict = {}


def set_const_vars_map(vars_map: dict) -> None:
    global _const_vars_map
    global _default_vars_map
    _const_vars_map = vars_map
    _default_vars_map = vars_map


def set_global_vars_map(vars_map: dict) -> None:
    global _global_vars_map
    global _default_vars_map
    _global_vars_map = vars_map
    _default_vars_map = copy.deepcopy(vars_map)
    _default_vars_map.update(_const_vars_map)


def dump_vars(vars_map: dict) -> None:
    start_raw_logging()
    for k, v in vars_map.items():
        verbose("{}='{}'", k, v)
    stop_raw_logging()


def dump_defualt_vars() -> None:
    dump_vars(_default_vars_map)


class StringVarExpander(object):
    var_re = re.compile(r'{{\S*?}}')

    def __init__(self, vars_map: Optional[dict] = None) -> None:
        if not vars_map:
            self.vars_map = _default_vars_map
        else:
            self.vars_map = copy.deepcopy(_global_vars_map)
            self.vars_map.update(vars_map)
            self.vars_map.update(_const_vars_map)

        self.expansion_stack = []

    def __call__(self, s: str) -> str:
        if self.expansion_stack:
            raise TaskException("Incomplete variable expansion?")
        return re.sub(StringVarExpander.var_re, self._expand_re, s)

    def _expand_re(self, match) -> str:
        var = match.group()[2:-2]
        if var in self.expansion_stack:
            raise TaskException(f"Recursive expanded var '{var}'")
        if var.startswith("$"):
            return os.getenv(var[1:], "")
        value = self.vars_map.get(var, "")
        if type(value) is list or type(value) is dict:
            raise TaskException(f"Var expanded path '{var}' doesn't refer to valid type")
        self.expansion_stack.append(var)
        s = re.sub(StringVarExpander.var_re, self._expand_re, str(value))
        self.expansion_stack.pop()
        return s


def parse_assignment_str(s: str) -> typing.Tuple[str, str]:
    parts = s.split('=', maxsplit=1)
    if len(parts) == 1:
        return s, ""
    return parts[0], parts[1]


def bt() -> None:
    traceback.print_stack(file=sys.stdout)


def print_dict(d: dict) -> None:
    print(json.dumps(d, indent=4))
